from typing import List

from matplotlib.patches import Arc

from mpl_format.patches import PatchListFormatter


class ArcListFormatter(PatchListFormatter):

    _patches: List[Arc]

    def __init__(self, patches: List[Arc]):

        super().__init__(patches)

    def x_centers(self) -> List[float]:

        return [arc.get_center()[0] for arc in self._patches]

    def y_centers(self) -> List[float]:

        return [arc.get_center()[1] for arc in self._patches]

    def widths(self) -> List[float]:

        return [arc.width for arc in self._patches]

    def heights(self) -> List[float]:

        return [arc.height for arc in self._patches]

    def angles(self) -> List[float]:

        return [arc.angle for arc in self._patches]

    def theta_starts(self) -> List[float]:

        return [arc.theta1 for arc in self._patches]

    def theta_ends(self) -> List[float]:

        return [arc.theta2 for arc in self._patches]



from enum import Enum
from typing import Union


class FONT_VARIANT(Enum):

    normal = 0
    small_caps = 1

    def get_name(self) -> str:

        return {
            'normal': 'normal',
            'small_caps': 'small-caps'
        }[self.name]

    @staticmethod
    def get_font_variant(
            font_variant: Union[str, 'FONT_VARIANT']
    ) -> Union[str, int, float]:
        if font_variant and isinstance(font_variant, FONT_VARIANT):
            font_variant = font_variant.get_name()
        return font_variant



from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.patches import Patch
from matplotlib.path import Path
from seaborn import JointGrid, PairGrid
from typing import TypeVar, Tuple, Union, Dict, Callable, Iterable

from mpl_format.enums import FONT_SIZE, FONT_STRETCH, FONT_WEIGHT, FONT_STYLE, \
    FONT_VARIANT, CAP_STYLE, JOIN_STYLE, LINE_STYLE, ARROW_STYLE, \
    CONNECTION_STYLE
from mpl_format.enums.box_style import BoxStyleType


ArrowStyle = TypeVar('ArrowStyle', str, ARROW_STYLE)
ArrowStyleIterable = Iterable[ArrowStyle]

BoxStyleTypeIterable = Iterable[BoxStyleType]

CapStyle = TypeVar('CapStyle', str, CAP_STYLE)
CapStyleIterable = Iterable[CapStyle]

Color = TypeVar(
    'Color',
    str,
    Tuple[float, float, float],
    Tuple[float, float, float, float]
)
ColorIterable = Iterable[Color]
ColorOrColorIterable = Union[Color, ColorIterable]

ConnectionStyle = TypeVar('ConnectionStyle', str, CONNECTION_STYLE)
ConnectionStyleIterable = Iterable[ConnectionStyle]

FontSize = TypeVar('FontSize', str, float, int, FONT_SIZE)
FontSizeIterable = Iterable[FontSize]

FontStretch = TypeVar('FontStretch', float, int, FONT_STRETCH)
FontStretchIterable = Iterable[FontStretch]

FontStyle = TypeVar('FontStyle', str, FONT_STYLE)
FontStyleIterable = Iterable[FontStyle]

FontVariant = TypeVar('FontVariant', str, FONT_VARIANT)
FontVariantIterable = Iterable[FontVariant]

FontWeight = TypeVar('FontWeight', int, float, str, FONT_WEIGHT)

JoinStyle = TypeVar('JoinStyle', str, JOIN_STYLE)
JoinStyleIterable = Iterable[JoinStyle]

LegendLocation = TypeVar('LegendLocation', str, Tuple[float, float])

LineStyle = TypeVar('LineStyle', str, LINE_STYLE)
LineStyleIterable = Iterable[LineStyle]

PatchIterable = Iterable[Patch]

PathIterable = Iterable[Path]

PlotObject = TypeVar('PlotObject', Axes, Figure, JointGrid, PairGrid)

StringMapper = Union[Dict[str, str], Callable[[str], str]]



from mpl_format.enums.arrow_style import ARROW_STYLE
from mpl_format.enums.box_style import BOX_STYLE
from mpl_format.enums.cap_style import CAP_STYLE
from mpl_format.enums.connection_style import CONNECTION_STYLE
from mpl_format.enums.draw_style import DRAW_STYLE
from mpl_format.enums.font_size import FONT_SIZE
from mpl_format.enums.font_stretch import FONT_STRETCH
from mpl_format.enums.font_style import FONT_STYLE
from mpl_format.enums.font_variant import FONT_VARIANT
from mpl_format.enums.font_weight import FONT_WEIGHT
from mpl_format.enums.join_style import JOIN_STYLE
from mpl_format.enums.line_style import LINE_STYLE
from mpl_format.enums.marker_style import MARKER_STYLE



from typing import Dict, Callable

from mpl_format.enums import \
    FONT_SIZE, FONT_STRETCH, FONT_STYLE, FONT_VARIANT, FONT_WEIGHT, \
    BOX_STYLE, CAP_STYLE, JOIN_STYLE, LINE_STYLE


kwarg_mappings: Dict[str, Callable] = {
    'boxstyle': BOX_STYLE.get_box_style,
    'capstyle': CAP_STYLE.get_cap_style,
    'fontsize': FONT_SIZE.get_font_size,
    'fontstretch': FONT_STRETCH.get_font_stretch,
    'fontstyle': FONT_STYLE.get_font_style,
    'fontvariant': FONT_VARIANT.get_font_variant,
    'fontweight': FONT_WEIGHT.get_font_weight,
    'joinstyle': JOIN_STYLE.get_join_style,
    'linestyle': LINE_STYLE.get_line_style
}



from enum import Enum
from typing import Union

from mpl_format.enums.connection_style import CONNECTION_STYLE


class DRAW_STYLE(Enum):

    default = 0
    steps = 1
    steps_pre = 2
    steps_mid = 3
    steps_post = 4

    def get_name(self) -> str:

        return {
            'default': 'default',
            'steps': 'steps',
            'steps_pre': 'steps-pre',
            'steps_mid': 'steps-mid',
            'steps_post': 'steps-post'
        }[self.name]

    @staticmethod
    def get_draw_style(
            draw_style: Union[str, 'DRAW_STYLE']
    ) -> str:
        if draw_style and isinstance(draw_style, CONNECTION_STYLE):
            draw_style = draw_style.get_name()
        return draw_style



from itertools import product
from typing import Union, List, Tuple, Iterator, Iterable

from matplotlib.axes import Axes
from matplotlib.axis import Axis

from compound_types.built_ins import FloatIterable
from mpl_format.compound_types import Color, FontSize, StringMapper
from mpl_format.enums.line_style import LINE_STYLE
from mpl_format.text.text_utils import wrap_text, map_text


class TicksFormatter(object):

    def __init__(self, axis: str, which: str, axes: Axes):
        """
        Create a new TicksFormatter.

        :param axis: One of {'x', 'y', 'both'}
        :param which: One of {'major', 'minor', 'both'}
        :param axes: Axes to operate on.
        """
        self._axis: str = axis
        self._which = which
        self._axes: Axes = axes

    def reset(self) -> 'TicksFormatter':
        """
        Set all parameters to defaults.
        """
        self._axes.tick_params(
            axis=self._axis,
            which=self._which,
            reset=True
        )
        return self

    @property
    def _is_minor(self) -> bool:
        return self._which == 'minor'

    # region values and labels

    def set_values(self, values: FloatIterable) -> 'TicksFormatter':
        """
        Set the values of the ticks.

        :param values: List of values where the ticks should be located.
        """
        x_axis = self._axes.xaxis
        y_axis = self._axes.yaxis
        if self._axis in ('x', 'both'):
            x_axis.set_ticks(ticks=values, minor=self._is_minor)
        if self._axis in ('y', 'both'):
            y_axis.set_ticks(ticks=values, minor=self._is_minor)
        return self

    def set_labels(self, labels: Iterable[str]) -> 'TicksFormatter':
        """
        Set the labels for the ticks.

        :param labels: List of labels to annotate each tick value.
        """
        x_axis = self._axes.xaxis
        y_axis = self._axes.yaxis
        if self._axis in ('x', 'both'):
            x_axis.set_ticklabels(ticklabels=labels, minor=self._is_minor)
        if self._axis in ('y', 'both'):
            y_axis.set_ticklabels(ticklabels=labels, minor=self._is_minor)
        return self

    def get_labels(self):

        raise NotImplementedError

    # endregion

    # region direction

    def set_direction(self, direction: str) -> 'TicksFormatter':
        """
        Puts ticks inside the axes, outside the axes, or both.

        :param direction: One of {'in', 'out', 'inout', 'in_out'}
        """
        direction = {
            'in': 'in',
            'out': 'out',
            'inout': 'inout',
            'in_out': 'inout'
        }[direction]
        self._axes.tick_params(axis=self._axis, which=self._which,
                               direction=direction)
        return self

    def set_direction_in(self) -> 'TicksFormatter':
        """
        Puts ticks inside the axes.
        """
        self.set_direction('in')
        return self

    def set_direction_out(self) -> 'TicksFormatter':
        """
        Puts ticks outside the axes.
        """
        self.set_direction('out')
        return self

    def set_direction_in_out(self) -> 'TicksFormatter':
        """
        Puts ticks inside and outside the axes.
        """
        self.set_direction('in_out')
        return self

    # endregion

    def set_length(self, length: float) -> 'TicksFormatter':
        """
        Set the tick length in points.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               length=length)
        return self

    def set_width(self, width: float) -> 'TicksFormatter':
        """
        Set the tick width in points.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               width=width)
        return self

    def set_padding(self, padding: float) -> 'TicksFormatter':
        """
        Set the distance in points between tick and label.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               pad=padding)
        return self

    def set_color(self, color: Color) -> 'TicksFormatter':
        """
        Set the tick color.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               color=color)
        return self

    def set_colors(self, color: Color) -> 'TicksFormatter':
        """
        Set tick color and the label color to the same value.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               colors=color)
        return self

    def set_z_order(self, z_order: float) -> 'TicksFormatter':
        """
        Set the tick and label z-order.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               zorder=z_order)
        return self

    # region show ticks

    def show_top(self, show: bool = True) -> 'TicksFormatter':
        """
        Whether to draw ticks at the top of the plot.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               top=show)
        return self

    def show_bottom(self, show: bool = True) -> 'TicksFormatter':
        """
        Whether to draw ticks at the bottom of the plot.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               bottom=show)
        return self

    def show_left(self, show: bool = True) -> 'TicksFormatter':
        """
        Whether to draw ticks at the left of the plot.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               left=show)
        return self

    def show_right(self, show: bool = True) -> 'TicksFormatter':
        """
        Whether to draw ticks at the right of the plot.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               right=show)
        return self

    # endregion

    # region labels

    def set_label_size(self, label_size: FontSize) -> 'TicksFormatter':
        """
        Set the tick label font size in points or as a string (e.g., 'large').
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               labelsize=label_size)
        return self

    def set_label_color(self, color: Color) -> 'TicksFormatter':
        """
        Set the tick label color.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               labelcolor=color)
        return self

    def set_label_rotation(self, rotation: float) -> 'TicksFormatter':
        """
        Set the tick label rotation.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               rotation=rotation)
        return self

    def _iter_axis_minor(self) -> Iterator[Tuple[Axis, bool]]:

        axes: List[Axis] = []
        axis: Axis
        if self._axis in ('x', 'both'):
            axes.append(self._axes.xaxis)
        if self._axis in ('y', 'both'):
            axes.append(self._axes.yaxis)
        minors: List[bool] = []
        if self._which in ('minor', 'both'):
            minors.append(True)
        if self._which in ('major', 'both'):
            minors.append(False)
        for axis, minor in product(axes, minors):
            yield axis, minor

    def wrap_label_text(self, max_width: int) -> 'TicksFormatter':
        """
        Wrap the text for each tick label with new lines if it exceeds
        a given width of characters.

        :param max_width: The maximum character width per line.
        """
        for axis, minor in self._iter_axis_minor():
            tick_labels = axis.get_ticklabels(minor=minor)
            if all(t.get_text() == '' for t in tick_labels):
                continue  # non categorical tick-labels e.g. line plot
            axis.set_ticklabels(
                ticklabels=[wrap_text(text, max_width)
                            for text in tick_labels],
                minor=minor
            )
        return self

    def map_label_text(self, mapping: StringMapper) -> 'TicksFormatter':
        """
        Map the tick label text using a dictionary or function.

        :param mapping: Dictionary or a function mapping old text to new text.
        """
        for axis, minor in self._iter_axis_minor():
            labels = [label.get_text()
                      for label in axis.get_ticklabels(minor=minor)]
            axis.set_ticklabels(
                ticklabels=map_text(labels, mapping),
                minor=minor
            )
        return self

    # region show labels

    def show_top_labels(self, show: bool = True) -> 'TicksFormatter':
        """
        Whether to draw tick labels at the top of the plot.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               labeltop=show)
        return self

    def show_bottom_labels(self, show: bool = True) -> 'TicksFormatter':
        """
        Whether to draw tick labels at the bottom of the plot.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               labelbottom=show)
        return self

    def show_left_labels(self, show: bool = True) -> 'TicksFormatter':
        """
        Whether to draw tick labels at the left of the plot.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               labelleft=show)
        return self

    def show_right_labels(self, show: bool = True) -> 'TicksFormatter':
        """
        Whether to draw tick labels at the right of the plot.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               labelright=show)
        return self

    # endregion

    # endregion

    # region grid

    def set_grid_color(self, color: Color) -> 'TicksFormatter':
        """
        Set the grid line color.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               grid_color=color)
        return self

    def set_grid_alpha(self, alpha: float) -> 'TicksFormatter':
        """
        Set the transparency of grid lines: 0 (transparent) to 1 (opaque).
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               grid_alpha=alpha)
        return self

    def set_grid_line_width(self, line_width: float) -> 'TicksFormatter':
        """
        Set the width of grid lines in points.
        """
        self._axes.tick_params(axis=self._axis, which=self._which,
                               grid_linewidth=line_width)
        return self

    def set_grid_line_style(self, line_style: Union[str, LINE_STYLE]):
        """
        Set the line style of the grid lines.
        """
        self._axes.tick_params(
            axis=self._axis, which=self._which,
            grid_linestyle=LINE_STYLE.get_line_style(line_style)
        )
        return self

    # endregion



from typing import Union, List

from celluloid import Camera
from numpy import linspace

from mpl_format.animation.shapes.base import ShapeAnimation
from mpl_format.axes import AxesFormatter
from mpl_format.figures import FigureFormatter


class AxesAnimator(object):

    def __init__(self,
                 formatter: Union[AxesFormatter, FigureFormatter],
                 duration: float):

        self.formatter: AxesFormatter = formatter or AxesFormatter()
        self.duration: float = duration
        self.shapes: List[ShapeAnimation] = []

    def add_animation(self, animation: ShapeAnimation):

        self.shapes.append(animation)

    def animate(self, file_name: str, fps: float = 30):

        camera = Camera(self.formatter.axes.figure)
        for t in linspace(
                0, self.duration, 1 + int(self.duration * fps)
        ):
            for shape in self.shapes:
                shape.draw(t / self.duration, axes=self.formatter)
            camera.snap()
        animation = camera.animate()
        animation.save(file_name, writer='imagemagick', fps=30)



from compound_types.built_ins import Scalar


def format_as_integer(number: Scalar, kmbt: bool = False) -> str:
    """
    Format a number as an integer, with comma separators and an optional suffix
    of K, M, B or T for large numbers.

    :param number: The number to format.
    :param kmbt: Whether to abbreviate numbers using K, M, B and T for
                 thousands, millions, billions and trillions.
    :return:
    """
    if not kmbt:
        return f'{int(number):,}'
    else:
        for power, abbrev in zip(
                [12, 9, 6, 3],
                ['T', 'B', 'M', 'K']
        ):
            if number >= 10 ** power:
                num = number / 10 ** power
                if num == int(num):
                    num = int(num)
                return f'{num:,}{abbrev}'
        if number == int(number):
            number = int(number)
        return f'{number:,}'



from typing import Sized, Dict, Any, Callable


def smart_zip(*args):
    """
    Method to convert arguments into a zipped list.
    Similar to Python's built-in zip method but converts each non-Sized into
    a list so that it can be iterated over by the user without explicitly
    checking its type.

    :param args: Each arg should be a single value or a Sized of values. Sized
                 args which are shorter than the longest Sized will be repeated
                len(Sized) times like any non-Sized arg. Dicts are treated as
                non-Sizeds.
    """
    max_arg_length = 1
    values = []
    # find longest sized arg
    for arg in args:
        if (
                isinstance(arg, Sized) and
                not isinstance(arg, dict) and
                not isinstance(arg, str)
        ):
            arg_length = len(arg)
            if arg_length > max_arg_length:
                max_arg_length = arg_length
    # create values
    for arg in args:
        if (
                isinstance(arg, Sized) and
                not isinstance(arg, dict) and
                not isinstance(arg, str) and
                len(arg) == max_arg_length
        ):
            values.append(arg)
        else:
            values.append([arg] * max_arg_length)
    for value_set in zip(*values):
        yield value_set


def smart_zip_kwargs(**kwargs):
    """
    Takes kwargs, passes the args into smart_zip and yields dicts mapping kws to
    zipped arg values.

    :param kwargs: Keys and values. Values passed into smart_zip.
    :return:
    """
    keys = list(kwargs.keys())
    values = list(kwargs.values())
    for value_set in smart_zip(*values):
        out_dict = {
            key: value for key, value in zip(keys, value_set)
        }
        yield out_dict


def drop_none_values(items: dict) -> dict:
    """
    Return a copy of the dictionary without any keys or values where the value
    is None.
    """
    return {
        key: value for key, value in items.items()
        if value is not None
    }


def apply_mappings(items: Dict[str, Any],
                   mappings: Dict[str, Callable]) -> Dict[str, Any]:
    """
    Return a copy of the dictionary with any items with a key in mappings
    transformed by the callable value associated with that key.

    :param items: dict of items to transform
    :param mappings: dict of mappings.
    """
    return {
        key: mappings[key](value) if key in mappings.keys() else value
        for key, value in items.items()
    }



from enum import Enum
from typing import Union


class FONT_STYLE(Enum):

    normal = 0
    italic = 1
    oblique = 2

    @staticmethod
    def get_font_style(
            font_style: Union[int, float, 'FONT_STYLE']
    ) -> Union[str, int, float]:
        if font_style and isinstance(font_style, FONT_STYLE):
            font_style = font_style.get_name()
        return font_style



from typing import List, Union, Optional

from numpy import linspace, sin, pi, sign, arcsin

from mpl_format.compound_types import Color
from mpl_format.animation.rate import Rate
from mpl_format.utils.color_utils import set_alpha, cross_fade


class ColorAnimation(object):

    def __init__(self,
                 colors: List[Color],
                 t: Optional[List[float]] = None,
                 rate: Optional[Union[Rate, str]] = None):

        self.colors: List[Color] = colors
        if t is not None:
            self.t: List[float] = t
        else:
            self.t = list(linspace(0, 1, len(colors)))
        self.rate: Rate = (
            Rate.linear() if rate is None else
            rate if isinstance(rate, Rate) else
            Rate(rate)
        )

    @staticmethod
    def fade_in(color: Color) -> 'ColorAnimation':

        return ColorAnimation(
            colors=[set_alpha(color, 0), color],
            t=[0, 1]
        )

    @classmethod
    def sine_wave(cls, color_1: Color, color_2: Color,
                  cycles: float, phase: float = 0.0) -> 'ColorAnimation':
        """
        Return a sine wave oscillating between color_1 and color_2,
        starting at phase * 2π with cycles complete waves.

        :param color_1: Lowest color of the wave.
        :param color_2: Highest color of the wave.
        :param cycles: Number of complete oscillations.
        :param phase: Value from 0 to 1 of where to start the wave.
        """
        middle_val = 0.5
        amplitude = 0.5
        rate = Rate(lambda t: (
                middle_val +
                amplitude * sin((2 * pi * t * cycles) - (2 * pi * phase))
        ))
        return ColorAnimation(colors=[color_1, color_2],
                              rate=rate)

    @classmethod
    def square_wave(cls, color_1: Color, color_2: Color,
                    cycles: float, phase: float = 0.0) -> 'ColorAnimation':
        """
        Return a square wave oscillating between min_val and max_val,
        starting at phase * 2π with cycles complete waves.

        :param color_1: Lowest color of the wave.
        :param color_2: Highest color of the wave.
        :param cycles: Number of complete oscillations.
        :param phase: Value from 0 to 1 of where to start the wave.
        """
        middle_val = 0.5
        amplitude = 0.5
        rate = Rate(lambda t: (
                middle_val +
                amplitude * sign(
                    sin((2 * pi * t * cycles) - (2 * pi * phase))
                )
        ))
        return ColorAnimation(colors=[color_1, color_2],
                              rate=rate)

    @classmethod
    def triangle_wave(cls, color_1: Color, color_2: Color,
                      cycles: float, phase: float = 0.0) -> 'ColorAnimation':
        """
        Return a triangle wave oscillating between min_val and max_val,
        starting at phase * 2π with cycles complete waves.

        :param color_1: Lowest color of the wave.
        :param color_2: Highest color of the wave.
        :param cycles: Number of complete oscillations.
        :param phase: Value from 0 to 1 of where to start the wave.
        """
        middle_val = 0.5
        amplitude = 0.5
        rate = Rate(lambda t: (
                middle_val +
                2 * amplitude * arcsin(
                    sin((2 * pi * t * cycles) - (2 * pi * phase))
                ) / pi
        ))
        return ColorAnimation(colors=[color_1, color_2],
                              rate=rate)

    def at(self, t: float) -> Color:

        for i in range(len(self.t)):
            if self.t[i + 1] >= t:
                dt = self.rate(
                    (t - self.t[i]) /
                    (self.t[i + 1] - self.t[i])
                )
                return cross_fade(
                    from_color=self.colors[i],
                    to_color=self.colors[i + 1],
                    amount=dt
                )



from typing import Union, Optional

from numpy import linspace
from scipy.interpolate import interp1d

from compound_types.built_ins import FloatIterable
from mpl_format.animation.kwarg_animations.float_animation import FloatAnimation
from mpl_format.animation.shapes.base import ShapeAnimation
from mpl_format.animation.type_animations import \
    FloatOrFloatAnimation, ColorOrColorAnimation, StrOrFloatAnimation, \
    StrOrFloatOrFloatAnimation
from mpl_format.axes import AxesFormatter
from mpl_format.enums.marker_style import MARKER_STYLE
from mpl_format.enums.draw_style import DRAW_STYLE
from mpl_format.enums.line_style import LINE_STYLE


class LineAnimation(ShapeAnimation, object):

    def __init__(
            self,
            x: FloatIterable,
            y: FloatIterable,
            smooth: Union[bool, int] = False,
            smooth_order: int = 2,
            length: Optional[StrOrFloatAnimation] = None,
            alpha: Optional[StrOrFloatOrFloatAnimation] = None,
            color: Optional[ColorOrColorAnimation] = None,
            draw_style: Optional[Union[str, DRAW_STYLE]] = None,
            label: Optional[str] = None,
            line_style: Optional[Union[str, LINE_STYLE]] = None,
            line_width: Optional[FloatOrFloatAnimation] = None,
            marker: Optional[Union[str, MARKER_STYLE]] = None,
            marker_edge_color: Optional[ColorOrColorAnimation] = None,
            marker_edge_width: Optional[FloatOrFloatAnimation] = None,
            marker_face_color: Optional[ColorOrColorAnimation] = None,
            marker_face_color_alt: Optional[ColorOrColorAnimation] = None,
            marker_size: Optional[FloatOrFloatAnimation] = None
    ):
        if smooth is not False:
            if smooth is True:
                smooth = 1000
            f_smooth = interp1d(x, y, kind=smooth_order)
            x_smooth = linspace(min(x), max(x), smooth)
            y_smooth = f_smooth(x_smooth)
            x = x_smooth
            y = y_smooth
        self.x: FloatIterable = x
        self.y: FloatIterable = y
        self.length: Optional[FloatAnimation] = self._float_anim(length)
        self.alpha: FloatOrFloatAnimation = self._float_anim(alpha)
        self.color: Optional[ColorOrColorAnimation] = color
        self.draw_style: Optional[Union[str, DRAW_STYLE]] = draw_style
        self.label: Optional[str] = label
        self.line_style: Optional[Union[str, LINE_STYLE]] = line_style
        self.line_width: Optional[FloatOrFloatAnimation] = line_width
        self.marker: Optional[Union[str, MARKER_STYLE]] = marker
        self.marker_edge_color: Optional[ColorOrColorAnimation] = \
            marker_edge_color
        self.marker_edge_width: Optional[FloatOrFloatAnimation] = \
            marker_edge_width
        self.marker_face_color: Optional[ColorOrColorAnimation] = \
            marker_face_color
        self.marker_face_color_alt: Optional[ColorOrColorAnimation] = \
            marker_face_color_alt
        self.marker_size: Optional[FloatOrFloatAnimation] = marker_size

    def draw(self, t: float, axes: AxesFormatter):

        kwargs = {}

        if self.length is not None:
            kwargs['x'] = self.x[: int(round(self.length.at(t) * len(self.x)))]
            kwargs['y'] = self.y[: int(round(self.length.at(t) * len(self.y)))]
        else:
            kwargs['x'] = self.x
            kwargs['y'] = self.y
        for float_kwarg in ('alpha', 'line_width',
                            'marker_edge_width', 'marker_size'):
            self._add_float_kwarg(float_kwarg, kwargs, t)
        for color_kwarg in ('color', 'marker_edge_color',
                            'marker_face_color', 'marker_face_color_alt'):
            self._add_color_kwarg(color_kwarg, kwargs, t)
        for kwarg in ('draw_style', 'label', 'line_style', 'marker'):
            self._add_non_animated_kwarg(kwarg, kwargs)

        axes.add_line(**kwargs)



from typing import Union

from mpl_format.animation.kwarg_animations.color_animation import ColorAnimation
from mpl_format.animation.kwarg_animations.float_animation import FloatAnimation
from mpl_format.compound_types import Color

FloatOrFloatAnimation = Union[float, FloatAnimation]
StrOrFloatAnimation = Union[str, FloatAnimation]
StrOrFloatOrFloatAnimation = Union[str, float, FloatAnimation]
ColorOrColorAnimation = Union[Color, ColorAnimation]



from mpl_format.axes.axes_formatter import AxesFormatter
from mpl_format.axes.axis_formatter import AxisFormatter
from mpl_format.axes.axes_formatter_array import AxesFormatterArray
from mpl_format.axes.axis_formatter_array import AxisFormatterArray
from mpl_format.axes.ticks_formatter import TicksFormatter



from typing import Optional, Union

from mpl_format.animation.shapes.base import ShapeAnimation
from mpl_format.animation.type_animations import FloatOrFloatAnimation, \
    StrOrFloatOrFloatAnimation, ColorOrColorAnimation
from mpl_format.axes import AxesFormatter
from mpl_format.enums.join_style import JOIN_STYLE
from mpl_format.enums.cap_style import CAP_STYLE
from mpl_format.enums.line_style import LINE_STYLE


class EllipseAnimation(ShapeAnimation, object):

    def __init__(
            self,
            x_center: FloatOrFloatAnimation,
            y_center: FloatOrFloatAnimation,
            width: FloatOrFloatAnimation,
            height: FloatOrFloatAnimation,
            angle: FloatOrFloatAnimation = 0.0,
            alpha: Optional[StrOrFloatOrFloatAnimation] = None,
            cap_style: Optional[Union[str, CAP_STYLE]] = None,
            color: Optional[ColorOrColorAnimation] = None,
            edge_color: Optional[ColorOrColorAnimation] = None,
            face_color: Optional[ColorOrColorAnimation] = None,
            fill: bool = True,
            join_style: Optional[Union[str, JOIN_STYLE]] = None,
            label: Optional[str] = None,
            line_style: Optional[Union[str, LINE_STYLE]] = None,
            line_width: Optional[FloatOrFloatAnimation] = None,
    ):
        self.x_center: FloatOrFloatAnimation = x_center
        self.y_center: FloatOrFloatAnimation = y_center
        self.width: FloatOrFloatAnimation = width
        self.height: FloatOrFloatAnimation = height
        self.angle: FloatOrFloatAnimation = angle
        self.alpha: Optional[StrOrFloatOrFloatAnimation] = alpha
        self.cap_style: Optional[Union[str, CAP_STYLE]] = cap_style
        self.color: Optional[ColorOrColorAnimation] = color
        self.edge_color: Optional[ColorOrColorAnimation] = edge_color
        self.face_color: Optional[ColorOrColorAnimation] = face_color
        self.fill: bool = fill
        self.join_style: Optional[Union[str, JOIN_STYLE]] = join_style
        self.label: Optional[str] = label
        self.line_style: Optional[Union[str, LINE_STYLE]] = line_style
        self.line_width: Optional[FloatOrFloatAnimation] = line_width

    def draw(self, t: float, axes: AxesFormatter):

        kwargs = {}
        for float_kwarg in (
                'x_center', 'y_center',
                'width', 'height', 'angle',
                'alpha', 'line_width'
        ):
            self._add_float_kwarg(float_kwarg, kwargs, t)
        for color_kwarg in ('color', 'edge_color', 'face_color'):
            self._add_color_kwarg(color_kwarg, kwargs, t)
        for kwarg in ('cap_style', 'fill', 'join_style',
                      'label', 'line_style'):
            self._add_non_animated_kwarg(kwarg, kwargs)

        axes.add_ellipse(**kwargs)



from enum import Enum
from typing import Union

from matplotlib.patches import BoxStyle


class BOX_STYLE(Enum):

    circle = 0
    double_arrow = 1
    left_arrow = 2
    right_arrow = 3
    round = 4
    round_4 = 5
    round_tooth = 6
    saw_tooth = 7
    square = 8

    def get_name(self) -> str:

        return {
            'circle': 'circle',
            'double_arrow': 'darrow',
            'left_arrow': 'larrow',
            'right_arrow': 'rarrow',
            'round': 'round',
            'round_4': 'round4',
            'round_tooth': 'roundtooth',
            'saw_tooth': 'sawtooth',
            'square': 'square'
        }[self.name]

    @staticmethod
    def get_box_style(
            box_style: Union[str, 'BOX_STYLE']
    ) -> str:
        if box_style and isinstance(box_style, BOX_STYLE):
            box_style = box_style.get_name()
        return box_style


BoxStyleType = Union[
    str, BOX_STYLE,
    BoxStyle.Circle, BoxStyle.Round, BoxStyle.Round4,
    BoxStyle.DArrow, BoxStyle.LArrow, BoxStyle.RArrow,
    BoxStyle.Roundtooth, BoxStyle.Sawtooth, BoxStyle.Square
]



from enum import Enum
from typing import Union


class FONT_STRETCH(Enum):

    ultra_condensed = 0
    extra_condensed = 1
    condensed = 2
    semi_condensed = 3
    normal = 4
    semi_expanded = 5
    expanded = 6
    extra_expanded = 7
    ultra_expanded = 8

    def get_name(self) -> str:

        return {
            'ultra_condensed': 'ultra-condensed',
            'extra_condensed': 'extra-condensed',
            'condensed': 'condensed',
            'semi_condensed': 'semi-condensed',
            'normal': 'normal',
            'semi_expanded': 'semi-expanded',
            'expanded': 'expanded',
            'extra_expanded': 'extra-expanded',
            'ultra_expanded': 'ultra-expanded'
        }[self.name]

    @staticmethod
    def get_font_stretch(
            font_stretch: Union[int, float, 'FONT_STRETCH']
    ) -> Union[str, int, float]:
        if font_stretch and isinstance(font_stretch, FONT_STRETCH):
            font_stretch = font_stretch.get_name()
        return font_stretch



from typing import Optional

from mpl_format.animation.kwarg_animations.color_animation import ColorAnimation
from mpl_format.animation.kwarg_animations.float_animation import FloatAnimation
from mpl_format.animation.type_animations import StrOrFloatAnimation
from mpl_format.animation.rate import Rate
from mpl_format.axes import AxesFormatter


class ShapeAnimation(object):

    @staticmethod
    def _float_anim(
            variable: Optional[StrOrFloatAnimation]
    ) -> Optional[FloatAnimation]:

        if isinstance(variable, str):
            variable = FloatAnimation(rate=variable)
        return variable

    def draw(self, t: float, axes: AxesFormatter):

        raise NotImplementedError

    def _add_non_animated_kwarg(self, arg_name: str, kwargs: dict):

        value = getattr(self, arg_name)
        if value is not None:
            kwargs[arg_name] = value

    def _add_color_kwarg(self, arg_name: str, kwargs: dict, t: float):

        value = getattr(self, arg_name)
        if value is not None:
            kwargs[arg_name] = (
                value.at(t) if isinstance(value, ColorAnimation)
                else value
            )

    def _add_float_kwarg(self, arg_name: str, kwargs: dict, t: float):

        value = getattr(self, arg_name)
        if value is not None:
            kwargs[arg_name] = (
                value.at(t) if isinstance(value, FloatAnimation)
                else Rate(value)(t) if isinstance(value, str)
                else value
            )



from enum import Enum
from typing import Union


class ARROW_STYLE(Enum):

    Curve = 1
    CurveB = 2
    BracketB = 3
    CurveFilledB = 4
    CurveA = 5
    CurveAB = 6
    CurveFilledA = 7
    CurveFilledAB = 8
    BracketA = 9
    BracketAB = 10
    Fancy = 11
    Simple = 12
    Wedge = 13
    BarAB = 14

    def get_name(self) -> str:

        return {
            'Curve': '-',
            'CurveB': '->',
            'BracketB': '-[',
            'CurveFilledB': '-|>',
            'CurveA': '<-',
            'CurveAB': '<->',
            'CurveFilledA': '<|-',
            'CurveFilledAB': '<|-|>',
            'BracketA': ']-',
            'BracketAB': ']-[',
            'Fancy': 'fancy',
            'Simple': 'simple',
            'Wedge': 'wedge',
            'BarAB': '|-|',
        }[self.name]

    @staticmethod
    def get_arrow_style(
            arrow_style: Union[str, 'ARROW_STYLE']
    ) -> str:
        if arrow_style and isinstance(arrow_style, ARROW_STYLE):
            arrow_style = arrow_style.get_name()
        return arrow_style



import matplotlib.pyplot as plt
from typing import Optional, List, Union

from matplotlib.axes import Axes
from numpy import concatenate
from pandas import DataFrame
from scipy.stats import poisson

from mpl_format.axes import AxesFormatter
from mpl_format.axes.axis_utils import new_axes
from mpl_format.compound_types import Color


def categorical_discrete_values_histogram(
        data: DataFrame,
        x: str, y: str,
        categories: Optional[List[str]] = None,
        colors: Union[Color, List[Color]] = 'k',
        ax: Optional[Axes] = None
):
    """
    Plot a histogram of several categories together.

    :param data: DataFrame with columns x and y
    :param x: Name of the category column. Categories will be found from the
              unique values of this column.
    :param y: Name of the value column. Each value is assumed to be a count,
              and the number of each count value will be counted.
    :param categories: Optional override for categories of x. Use to select a
                       subset of x, or to reorder x.
    :param colors: Single color or list of colors to override default color of
                   each histogram.
    :param ax: Optional matplotlib Axes instance to plot on.
    """
    ax: Axes = ax or new_axes()
    axf = AxesFormatter(ax)
    y_min = data[y].min()
    y_max = data[y].max()
    z = data.groupby([x, y]).size()
    z_max = z.max()
    if categories is None:
        categories = data[x].unique()
    num_cats = len(categories)
    if not isinstance(colors, list):
        colors = [colors] * num_cats
    for c, category, color in zip(
        range(num_cats),
        categories,
        colors
    ):
        cat_hist = z.loc[category, :]
        for (_, y_i), z_i in cat_hist.items():
            axf.add_rectangle(
                width=0.8, height=1,
                x_center=c + 1, y_center=y_i,
                alpha=z_i / z_max,
                color=color
            )
    axf.set_x_lim(0.5)
    axf.set_x_max(len(categories) + 0.5)
    axf.set_y_min(y_min - 1)
    axf.set_y_max(y_max + 1)
    axf.x_ticks.set_values(list(range(1, 1 + len(categories))))
    axf.x_ticks.set_labels(categories)
    axf.set_text(
        x_label=x, y_label=y
    )
    return ax


def do_test_plot():

    p = concatenate([
        poisson(3).rvs(1_000),
        poisson(7).rvs(1_000),
        poisson(10).rvs(1_000)
    ])
    d = DataFrame({
        'y': p,
        'x': ['a'] * 1000 + ['b'] * 1_000 + ['c'] * 1_000
    })
    ax = categorical_discrete_values_histogram(
        data=d, x='x', y='y'
    )
    plt.show()
    ax = categorical_discrete_values_histogram(
        data=d, x='x', y='y', categories=['c', 'a']
    )
    plt.show()
    ax = categorical_discrete_values_histogram(
        data=d, x='x', y='y', categories=['c', 'a'], colors=['r', 'g']
    )
    AxesFormatter(ax).set_axis_below().grid()
    plt.show()


if __name__ == '__main__':

    do_test_plot()



from enum import Enum
from typing import Optional, Union


class JOIN_STYLE(Enum):

    miter = 1
    round = 2
    bevel = 3

    @staticmethod
    def get_join_style(
            join_style: Optional[Union[str, 'JOIN_STYLE']] = None
    ) -> str:
        if join_style and isinstance(join_style, JOIN_STYLE):
            join_style = join_style.name
        return join_style



from .arc_animation import ArcAnimation
from .arrow_animation import ArrowAnimation
from .circle_animation import CircleAnimation
from .ellipse_animation import EllipseAnimation
from .line_animation import LineAnimation
from .rectangle_animation import RectangleAnimation
from .regular_polygon_animation import RegularPolygonAnimation
from .wedge_animation import WedgeAnimation



from enum import Enum
from typing import Optional, Union


class LINE_STYLE(Enum):

    solid = 1
    dashed = 2
    dash_dot = 3
    dotted = 4

    def get_name(self) -> str:

        return {
            'solid': 'solid',
            'dashed': 'dashed',
            'dash_dot': 'dashdot',
            'dotted': 'dotted',
        }[self.name]

    @staticmethod
    def get_line_style(
            line_style: Optional[Union[str, 'LINE_STYLE']] = None) -> str:
        if line_style and isinstance(line_style, LINE_STYLE):
            line_style = line_style.name
        return line_style



from numpy import pi, sin, sign, arcsin
from typing import List, Union, Optional

from mpl_format.animation.rate import Rate


class FloatAnimation(object):

    def __init__(self,
                 values: Optional[List[float]] = None,
                 t: Optional[List[float]] = None,
                 rate: Optional[Union[Rate, str]] = None):

        if values is not None:
            self.values: List[float] = values
        else:
            self.values = [0, 1]
        if t is not None:
            self.t: List[float] = t
        else:
            self.t = [0, 1]
        self.rate: Rate = (
            Rate.linear() if rate is None else
            rate if isinstance(rate, Rate) else
            Rate(rate)
        )

    def set_values(self, values: List[float]) -> 'FloatAnimation':

        self.values = values
        return self

    def reverse_values(self) -> 'FloatAnimation':

        return FloatAnimation(
            values=self.values[:: -1],
            t=self.t, rate=self.rate
        )

    @classmethod
    def linear(cls) -> 'FloatAnimation':
        return FloatAnimation(rate='linear')

    @classmethod
    def quadratic(cls) -> 'FloatAnimation':
        return FloatAnimation(rate='quadratic')

    @classmethod
    def cubic(cls) -> 'FloatAnimation':
        return FloatAnimation(rate='cubic')

    @classmethod
    def sine_wave(cls, min_val: float, max_val: float,
                  cycles: float, phase: float = 0.0) -> 'FloatAnimation':
        """
        Return a sine wave oscillating between min_val and max_val,
        starting at phase * 2π with cycles complete waves.

        :param min_val: Lowest value of the wave.
        :param max_val: Highest value of the wave.
        :param cycles: Number of complete oscillations.
        :param phase: Value from 0 to 1 of where to start the wave.
        """
        middle_val = (min_val + max_val) / 2
        amplitude = (max_val - min_val) / 2
        rate = Rate(lambda t: (
                middle_val +
                amplitude * sin((2 * pi * t * cycles) - (2 * pi * phase))
        ))
        return FloatAnimation(rate=rate)

    @classmethod
    def square_wave(cls, min_val: float, max_val: float,
                    cycles: float, phase: float = 0.0) -> 'FloatAnimation':
        """
        Return a square wave oscillating between min_val and max_val,
        starting at phase * 2π with cycles complete waves.

        :param min_val: Lowest value of the wave.
        :param max_val: Highest value of the wave.
        :param cycles: Number of complete oscillations.
        :param phase: Value from 0 to 1 of where to start the wave.
        """
        middle_val = (min_val + max_val) / 2
        amplitude = (max_val - min_val) / 2
        rate = Rate(lambda t: (
                middle_val +
                amplitude * sign(
                    sin((2 * pi * t * cycles) - (2 * pi * phase))
                )
        ))
        return FloatAnimation(rate=rate)

    @classmethod
    def triangle_wave(cls, min_val: float, max_val: float,
                      cycles: float, phase: float = 0.0) -> 'FloatAnimation':
        """
        Return a triangle wave oscillating between min_val and max_val,
        starting at phase * 2π with cycles complete waves.

        :param min_val: Lowest value of the wave.
        :param max_val: Highest value of the wave.
        :param cycles: Number of complete oscillations.
        :param phase: Value from 0 to 1 of where to start the wave.
        """
        middle_val = (min_val + max_val) / 2
        amplitude = (max_val - min_val) / 2
        rate = Rate(lambda t: (
                middle_val +
                2 * amplitude * arcsin(
                    sin((2 * pi * t * cycles) - (2 * pi * phase))
                ) / pi
        ))
        return FloatAnimation(rate=rate)

    @classmethod
    def rotation(cls, cycles: float,
                 clockwise: bool = True,
                 phase: Optional[float] = 0.0):
        """
        Return a rotation in degrees.

        :param cycles: Number of complete rotations.
        :param clockwise: Rotate clockwise if True.
        :param phase: Starting phase between 0.0 and 1.0.
        """
        if clockwise:
            multiplier = -1
        else:
            multiplier = 1
        return FloatAnimation(rate=Rate(
            lambda t: (phase * 360) + multiplier * t * cycles * 360
        ))

    def at(self, t: float) -> float:

        for i in range(len(self.t)):
            if self.t[i + 1] >= t:
                dt = self.rate(
                    (t - self.t[i]) /
                    (self.t[i + 1] - self.t[i])
                )
                return (
                    self.values[i] +
                    (self.values[i + 1] - self.values[i]) * dt
                )



from enum import Enum
from typing import Union


class FONT_SIZE(Enum):

    xx_small = 0
    x_small = 1
    small = 2
    medium = 3
    large = 4
    x_large = 5
    xx_large = 6

    def get_name(self) -> str:

        return {
            'xx_small': 'xx-small',
            'x_small': 'x-small',
            'small': 'small',
            'medium': 'medium',
            'large': 'large',
            'x_large': 'x-large',
            'xx_large': 'xx-large'
        }[self.name]

    @staticmethod
    def get_font_size(
            font_size: Union[int, float, 'FONT_SIZE']
    ) -> Union[str, int, float]:
        if font_size and isinstance(font_size, FONT_SIZE):
            font_size = font_size.get_name()
        return font_size



from enum import Enum
from typing import Union


class FONT_WEIGHT(Enum):

    ultra_light = 0
    light = 1
    normal = 2
    regular = 3
    book = 4
    medium = 5
    roman = 6
    semi_bold = 7
    demi_bold = 8
    demi = 9
    bold = 10
    heavy = 11
    extra_bold = 12
    black = 13

    def get_name(self) -> str:

        return {
            'ultra_light': 'ultralight',
            'light': 'light',
            'normal': 'normal',
            'regular': 'regular',
            'book': 'book',
            'medium': 'medium',
            'roman': 'roman',
            'semi_bold': 'semibold',
            'demi_bold': 'demibold',
            'demi': 'demi',
            'bold': 'bold',
            'heavy': 'heavy',
            'extra_bold': 'extra bold',
            'black': 'black'
        }[self.name]

    @staticmethod
    def get_font_weight(
            font_weight: Union[int, float, 'FONT_WEIGHT']
    ) -> Union[str, int, float]:
        if font_weight and isinstance(font_weight, FONT_WEIGHT):
            font_weight = font_weight.get_name()
        return font_weight



from enum import Enum
from typing import Optional, Union


class CAP_STYLE(Enum):

    butt = 1
    round = 2
    projecting = 3

    @staticmethod
    def get_cap_style(
            cap_style: Optional[Union[str, 'CAP_STYLE']] = None
    ) -> str:
        if cap_style and isinstance(cap_style, CAP_STYLE):
            cap_style = cap_style.name
        return cap_style



from enum import Enum
from typing import Union


class CONNECTION_STYLE(Enum):

    angle = 1
    angle_3 = 2
    arc = 3
    arc_3 = 4
    bar = 5

    def get_name(self) -> str:

        return {
            'angle': 'angle',
            'angle_3': 'angle3',
            'arc': 'arc',
            'arc_3': 'arc3',
            'bar': 'bar'
        }[self.name]

    @staticmethod
    def get_connection_style(
            connection_style: Union[str, 'ARROW_STYLE']
    ) -> str:
        if connection_style and isinstance(connection_style, CONNECTION_STYLE):
            connection_style = connection_style.get_name()
        return connection_style



from typing import Callable, Union


class Rate(object):

    def __init__(self, lambda_t: Union[str, Callable[[float], float]]):

        if isinstance(lambda_t, str):
            if hasattr(self, lambda_t):
                self._lambda_t: Union[str, Callable[[float], float]] = \
                    getattr(self, lambda_t)()
            else:
                raise ValueError(f'Rate class has no method {lambda_t}')
        else:
            self._lambda_t: Callable[[float], float] = lambda_t

    def __call__(self, t: float) -> float:

        return self._lambda_t(t)

    @classmethod
    def linear(cls) -> 'Rate':
        return Rate(lambda t: t)

    @classmethod
    def quadratic(cls) -> 'Rate':
        return Rate(lambda t: t ** 2)

    @classmethod
    def cubic(cls) -> 'Rate':
        return Rate(lambda t: t ** 3)



def check_h_align(h_align: str):

    if h_align not in ('left', 'center', 'right'):
        raise ValueError("h_align not in ('left', 'center', 'right')")


def check_v_align(h_align: str):

    if h_align not in ('top', 'center', 'bottom'):
        raise ValueError("v_align not in ('top', 'center', 'bottom')")



from typing import Optional, Union

from mpl_format.animation.kwarg_animations import FloatAnimation
from mpl_format.animation.shapes.base import ShapeAnimation
from mpl_format.animation.type_animations import FloatOrFloatAnimation, \
    StrOrFloatAnimation, ColorOrColorAnimation
from mpl_format.compound_types import Color
from mpl_format.enums import FONT_SIZE, FONT_STRETCH, FONT_STYLE, \
    FONT_VARIANT, FONT_WEIGHT, CAP_STYLE, JOIN_STYLE, LINE_STYLE


class TextAnimation(ShapeAnimation, object):

    def __init__(
            self, x: FloatOrFloatAnimation,
            y: FloatOrFloatAnimation,
            text: str,
            length: Optional[StrOrFloatAnimation] = None,
            font_dict: Optional[dict] = None,
            alpha: Optional[float] = None,
            color: Optional[ColorOrColorAnimation] = None,
            h_align: Optional[str] = None,
            v_align: Optional[str] = None,
            m_align: Optional[str] = None,
            line_spacing: Optional[float] = None,
            font_family: Optional[str] = None,
            font_size: Optional[Union[int, float, str, FONT_SIZE]] = None,
            font_stretch: Optional[Union[int, float, FONT_STRETCH]] = None,
            font_style: Optional[Union[str, FONT_STYLE]] = None,
            font_variant: Optional[Union[str, FONT_VARIANT]] = None,
            font_weight: Optional[Union[int, float, str, FONT_WEIGHT]] = None,
            wrap: Optional[bool] = None,
            bbox_style: Optional[str] = None,
            bbox_alpha: Optional[float] = None,
            bbox_cap_style: Optional[Union[str, CAP_STYLE]] = None,
            bbox_color: Optional[Color] = None,
            bbox_edge_color: Optional[Color] = None,
            bbox_face_color: Optional[Color] = None,
            bbox_fill: Optional[bool] = None,
            bbox_join_style: Optional[Union[str, JOIN_STYLE]] = None,
            bbox_line_style: Optional[Union[str, LINE_STYLE]] = None,
            bbox_line_width: Optional[float] = None
    ):
        self.x: FloatOrFloatAnimation = x
        self.y: FloatOrFloatAnimation = y
        self.text: str = text
        self.length: Optional[FloatAnimation] = self._float_anim(length)
        self.font_dict: Optional[dict] = font_dict
        self.alpha: FloatOrFloatAnimation = self._float_anim(alpha)
        self.color: Optional[ColorOrColorAnimation] = color
        self.h_align: Optional[str] = h_align
        self.v_align: Optional[str] = v_align
        self.m_align: Optional[str] = m_align
        self.line_spacing: Optional[float] = line_spacing
        self.font_family: Optional[str] = font_family
        self.font_size: Optional[Union[int, float, str, FONT_SIZE]] = font_size
        self.font_stretch: Optional[
            Union[int, float, FONT_STRETCH]] = font_stretch
        self.font_style: Optional[Union[str, FONT_STYLE]] = font_style
        self.font_variant: Optional[Union[str, FONT_VARIANT]] = font_variant
        self.font_weight: Optional[
            Union[int, float, str, FONT_WEIGHT]] = font_weight
        self.wrap: Optional[bool] = wrap
        self.bbox_style: Optional[str] = bbox_style
        self.bbox_alpha: Optional[float] = bbox_alpha
        self.bbox_cap_style: Optional[Union[str, CAP_STYLE]] = bbox_cap_style
        self.bbox_color: Optional[Color] = bbox_color
        self.bbox_edge_color: Optional[Color] = bbox_edge_color
        self.bbox_face_color: Optional[Color] = bbox_face_color
        self.bbox_fill: Optional[bool] = bbox_fill
        self.bbox_join_style: Optional[Union[str, JOIN_STYLE]] = bbox_join_style
        self.bbox_line_style: Optional[Union[str, LINE_STYLE]] = bbox_line_style
        self.bbox_line_width: Optional[float] = bbox_line_width

    # def draw(self, t: float, axes: AxesFormatter):
    #
    #     kwargs = {}
    #
    #     if self.length is not None:
    #         num_chars = int(len(self.text) * self.length.at(t))
    #         kwargs['text'] = self.text[: num_chars]
    #     else:
    #         kwargs['text'] = self.text
    #
    #     for float_kwarg in ('x', 'y', '')
